// ****************************************************************************
// Copyright (C) 2004  Alexandre Suaide
//
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// The Crate Controler Wiener-CC32 is not authorized for use as
// critical component in life support systems without the express
// written approval of ARW Elektronik Germany.
//
// Maintainer(s): Alexandre Suaide (suaide@if.usp.br)
//
// ****************************************************************************

#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/fs.h>

  #if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,15))
    #define KERNEL2615
  #endif
#ifdef KERNEL2615
  #include <linux/hrtimer.h>
#endif
#include "c111at-drv.h"
#include "common/common.c"


/***************************************************************/
#ifdef KERNEL26
int __init init_module(void)
#else
static int init_module(void)
#endif
{
  int i;

  if (register_chrdev(AC_MAJOR, CC_NAME, &cc_fops))
  {
    printk(" ERROR: unable to get major %i for " CC_NAME " device\n",AC_MAJOR);
    return -EIO;
  }
	
	
  for(i=0;i<MAX_CRATES;i++)
  {
    printk("===============================\nSetting up for crate # %d \n",i);
    cc_in_use[i] = 0;
    set_crate_number(i);
    camcl(CRATEONLINE);
    camcl(ZEROCRATE);
    camcl(CLEARCRATE);
    camcl(CRATEONLINE);
    camcl(DISABLE);
    camcl(CRATESTATUS);
  }
  printk("===============================\n");
  printk (CC_NAME ": initializing module \n");
  return 0;
}
/***************************************************************/
#ifdef KERNEL26
void __exit cleanup_module(void)
#else
static void cleanup_module(void)
#endif
{	
  unregister_chrdev(AC_MAJOR, CC_NAME);
  free_irq(CC_IRQ,NULL);
  release_region(GEO_port,0x20);
  printk(CC_NAME ": module removed\n");
}
/***************************************************************/
static int cc_open(struct inode *inode, struct file *file)
{
  int istat;
  int i;

  #ifndef KERNEL26
    printk(CC_NAME": Checking region and registering ports\n");
    init_module();
    if(check_region(GEO_port,0x20))
    {
      printk(CC_NAME": Unable to register CAMAC ports\n");
      return -EIO;
    }
  #endif
  
  printk(CC_NAME": Requesting region\n");
  if(!request_region(GEO_port,0x20,"C111aT"))
  {
    printk(CC_NAME": Unable to register CAMAC ports\n");
    return -EIO;
  }
  
  printk(CC_NAME": Requesting IRQ\n");
  
  #ifndef SA_INTERRUPT
    #define SA_INTERRUPT IRQF_DISABLED
  #endif
  
  istat = request_irq(CC_IRQ, (void*)cc_interrupt, SA_INTERRUPT, "C111aT" ,NULL);
  
  if (istat)
  {
    printk(CC_NAME": can't allocate IRQ%d for DSP camac\n",CC_IRQ);
    release_region(GEO_port,0x20);
    return -EIO;
  }
  
  for(i=0;i<MAX_CRATES;i++)
  {
    printk("===============================\nSetting up for crate # %d \n",i);
    cc_in_use[i] = 0;
    set_crate_number(i);
    camcl(CRATEONLINE);
    camcl(ZEROCRATE);
    camcl(CLEARCRATE);
    camcl(CRATEONLINE);
    camcl(DISABLE);
    camcl(CRATEOFFLINE);
    init_crate(i);
  }
  if(cc_in_use[0] == 1)
  {
    printk("===============================\n");
    printk("Crate number 0 set as current\n");
    set_crate_number(0);
  }
  printk("===============================\n");
  printk("CAMAC open OK \n");
  return 0;
}
/***************************************************************/
static void init_crate(int crate)
{
  if (crate>=MAX_CRATES)
  {
    return;
  }
  if(cc_in_use[crate]==1)
  {
    printk("Crate already in Use...\n");
    return;
  }
  printk("===============================\n");
  printk("Initializing crate number %d \n",crate);
  cc_in_use[crate] = 1;
  set_crate_number(crate);
  camcl(ZEROCRATE);
  camcl(CLEARCRATE);
  camcl(CRATEONLINE);
  return;
}
/***************************************************************/
static void close_crate(int crate)
{
  cc_in_use[crate] = 0;
  set_crate_number(crate);
  camcl(ZEROCRATE);
  camcl(CRATEOFFLINE);
  return;
}
/***************************************************************/
static int cc_release(struct inode *inode, struct file *file)
{
  unsigned int crate = MINOR(inode->i_rdev);
  release_region(GEO_port,0x20);
  free_irq(CC_IRQ,NULL);
  close_crate(crate);
  return 0;
}
/***************************************************************/
#ifdef KERNEL26
static irqreturn_t cc_interrupt(int irq, void *dev, struct pt_regs *regs)
#else
static void  cc_interrupt(int irq, void *dev, struct pt_regs *regs)
#endif
{
  read_event_now();
  #ifdef KERNEL26
    local_irq_enable();
    return IRQ_HANDLED;
  #else
   sti();    
   return;
  #endif
}
/***************************************************************/
static void camcl(int I)
{
   int bits, opt;
   unsigned char st;
   opt=(0x000f&I);
   bits=(0x0ff00&I)>>8;
   switch(opt)
   {
     case ZEROCRATE:
       outb(0,Crate_Z);
       printk("Z on crate \n");
       break;
     case CLEARCRATE:
       outb_p(0,Crate_C);
       printk("Clear on crate \n");
       break;
     case SWITCHLAM:
       if(I==ENABLE) 
			 {
			   camcl(CRATEONLINE);
				 printk("Enable LAM\n");
       }
			 if(I==DISABLE)
			 {
			   camcl(CRATEOFFLINE);
				 printk("Disable LAM\n");
       }
			 printk("COMMAND FOR LAM bits = %x S_mod_W = %x\n",bits,S_mod_W);
       outb(bits,S_mod_W);
       break;
     case CRATEONLINE:
       outb_p(0,Crate_on);
       bits=0x000B;
       outb_p(bits,S_mod_W);
       printk("Crate On-line, Clear Inhibit \n");
       break;
     case CRATEOFFLINE:
       bits=0x000A;
       outb_p(bits,S_mod_W);
       outb_p(0,Crate_off);
       printk("Crate Off-line, Set Inhibit \n");
       break;
     case CRATESTATUS:
       st=inb_p(R_S)&0x07;
       printk("Crate Status: %x \n",st);
     default:
       break;
   }
}
