{$A+,B-,D+,E+,F+,G+,I-,L+,N+,O+,R-,S-,V+,X+}
{$M 16384,0,655360}
unit CAM_TU16;
{********************************************************************
*                CC16 / PC 16 C++ routines                          *
*  ---------------------------------------------------------------- *
*   Unit contains :                                                 *
*     cam_adr         - Set I/O Address (BADR)                      *
*     getstatus     	- Get CC16 Status                             *
*     cam_cratecheck	- Test crate available                        *
*     set_crate     	- Set geographical crate Address              *
*     cam_controller_ini - Controller Initialization                *
*     cam_0         	- set CAMAC Command register to               *
*     cam_z         	- CAMAC Initialize (Z)                        *
*     cam_c         	- CAMAC Clear (C)                             *
*     cam_i         	- CAMAC Inhibit (I)                           *
*     cam_ci        	- CAMAC Clear + Inhibit (I) setzen            *
*     cam_irq       	- enable CAMAC LAM-Interrupt                  *
*     cam_noirq     	- disable CAMAC LAM-Interrupt                 *
*     cam_q         	- get CAMAC Q-response                        *
*     cam_x          	- get CAMAC X-response                        *
*     cam_nfa       	- set N,F,A start CAMAC cycle                 *
*     cam_nfa_read  	- set N,F,A and read 2 byte                   *
*     cam_nfaqx_read   - set N,F,A and read 2 byte with Q and X     *
*     cam_nfaqx_read24 - set N,F,A and read 3 byte with Q and X     *
*     cam_nfa_write 	 - set N,F,A and write 2 byte                 *
*     cam_nfaqx_write  - set N,F,A and write 2 byte with Q and X    *
*     cam_nfaqx_write24- set N,F,A and write 2 byte with Q and X    *
*     cam_lam       	- get LAM source                              *
*     cam_lamf      	- any LAM request                             *
*                                                                   *
*     badr := basic I/O address                                     *
*     copyright: A. Ruben, W-Ie-Ne-R, Plein & Baus GmbH  29.08.96   *
*********************************************************************}
interface

type datarr = array[0 .. 4096] of word;

     TWordRec = record
       LoWo,HiWo : word;
     end;


var BAdr00,BAdr02,BAdr04,BAdr06,BAdr08,BAdr0A,BAdr0C,BAdr0E : integer;
    BAdr10,BAdr12,BAdr14,BAdr16,BAdr18,BAdr1A,BAdr1C,BAdr1E : integer;

procedure CAM_Adr(BAD : word);
procedure CAM_Controller_Ini(BAD : word);
function  getstatus : word;
function  CAM_CrateCheck: boolean;
Procedure Set_Crate(CNr : word);
procedure CAM_0;
procedure CAM_Z;
procedure CAM_C;
procedure CAM_I;
procedure CAM_CI;
procedure CAM_IRQ;
procedure CAM_NoIRQ;
function  CAM_Q: Boolean;
function  CAM_X: Boolean;
procedure CAM_NFA(N,F,A : integer);
function  CAM_NFA_Read(N,F,A : integer): word;
procedure CAM_NFA_Write(N,F,A,Data : integer);
function  CAM_NFA_Read24(N,F,A : integer): longint;
procedure CAM_NFA_Write24(N,F,A:integer;Data : longint);
function  CAM_NFA_ReadPS(N,F,A : integer):word;
procedure CAM_NFAQX_READ(N,F,A : integer; var Data : word;var Q, X : boolean);
procedure CAM_NFAQX_Write(N,F,A : integer;data : word; var Q, X : boolean);
procedure CAM_NFAQX_READ24(N,F,A : integer; var Data : Longint;var Q, X : boolean);
procedure CAM_NFAQX_Write24(N,F,A:integer;Data : longint; var Q, X : boolean);
function  CAM_LAM(Station:integer): Boolean;
function  CAM_LAMF: Boolean;
procedure CAM_NFA_BlockRead(N,F,A,length : integer;var buffer:datarr);
procedure broadc_on;
procedure broadc_off;

Implementation
{***************************************************************************}

procedure CAM_Controller_Ini(BAD:word);
var cnr:integer;
begin
  cnr:=0;
  CAM_Adr(BAD);
  Set_Crate(cnr);
  CAM_Z;
end;
{***************************************************************************}

procedure CAM_Adr(BAD : word);
begin
  BAdr00:=BAD;    { geogr. address     }
  BAdr02:=BAd+2;  { NFA + cycle bit    }
  BAdr04:=BAd+4;  { write high byte    }
  BAdr06:=BAd+6;  { write low word     }
  BAdr08:=BAd+8;  { read high byte     }
  BAdr0A:=BAd+10; { read low word      }
  BAdr0C:=BAd+12; { block read         }
  BAdr0E:=BAd+14; { write modes  I,C,Z }
  BAdr10:=BAD+16; { read modes I,On,Lam}
  BAdr12:=BAd+18; { read Q,X,LAM       }
  BAdr14:=BAd+20; { crate C cycle      }
  BAdr16:=BAd+22; { crate Z cycle      }
  BAdr18:=BAd+24; { set crate bus on   }
  BAdr1A:=BAd+26; { set crate bus off  }
  BAdr1C:=BAd+28; { free               }
  BAdr1E:=BAd+30; { free               }
end;
{***************************************************************************}

function getstatus : word;
var i : byte;
    j : integer;
begin
  j:=0;
  J:=port[badr10];
  getstatus:=j and 7;
end;
{***************************************************************************}

function CAM_CrateCheck: boolean;
var l : integer;
    ccrate : byte;
    present : boolean;
begin
  ccrate:=0;
  present:=true;
  portw[BAdr00]:=(16 or ccrate);
  CAM_Z;
  if ((not CAM_Q) or (not CAM_X)) then present:=false;
  CAM_CrateCheck:=present;
end;
{***************************************************************************}

Procedure Set_Crate(CNr : word);
begin
  portw[BAdr00]:=(16 or cnr);
end;
{***************************************************************************}

procedure CAM_Z;
begin
  portw[badr16]:=0;
end;
{***************************************************************************}

procedure CAM_C;
begin
  portw[badr14]:=0;
end;
{***************************************************************************}

procedure CAM_I;
begin
  portw[badr0E]:=10;
end;
{***************************************************************************}

procedure CAM_CI;
var i: byte;
begin
  CAM_C;
  CAM_I;
end;
{***************************************************************************}

procedure CAM_0;
var i: byte;
begin
  portw[badr0E]:=11;
end;
{***************************************************************************}

procedure CAM_IRQ;
begin
  portw[badr0E]:=12;
end;
{***************************************************************************}

procedure CAM_NoIRQ;
begin
  portw[badr0E]:=13;
end;
{***************************************************************************}

function CAM_Q: Boolean;
var i:byte;
begin
  i:=port[badr12] and 1;
  if i=1 then CAM_Q:=true else CAM_Q:=false;
end;
{***************************************************************************}

function CAM_X: Boolean;
var i: byte;
begin
  i:=port[badr12] and 2;
  if i=2 then CAM_X:=true else CAM_X:=false;
end;
{***************************************************************************}

procedure CAM_NFA(N,F,A : integer);
var naf : integer;
begin
	naf:=(A and $0F) shl 5;
  naf:=(naf or (F and $1F)) shl 5;
  naf:=naf or (N and $1F);
  portw[badr02]:=naf or $8000;
end;
{***************************************************************************}

function CAM_NFA_Read(N,F,A : integer):word;
var naf : integer;
begin
	naf:=(A and $0F) shl 5;
  naf:=(naf or (F and $1F)) shl 5;
  naf:=naf or (N and $1F);
  portw[badr02]:=naf or $8000;
  CAM_NFA_Read:=portw[badr0A];
end;
{***************************************************************************}

function CAM_NFA_ReadPS(N,F,A : integer):word;
var naf : integer;
begin
	naf:=(A and $0F) shl 5;
  naf:=(naf or (F and $1F)) shl 5;
  naf:=naf or (N and $1F);
  portw[badr02]:=naf or $8000;
  CAM_NFA_ReadPS:=portw[badr0A];
end;
{***************************************************************************}

procedure CAM_NFA_Write(N,F,A,Data : integer);
var naf : integer;
begin
	naf:=(A and $0F) shl 5;
  naf:=(naf or (F and $1F)) shl 5;
  naf:=naf or (N and $1F);
  portw[badr02]:=naf;
  portw[badr06]:=data;
end;
{***************************************************************************}

procedure CAM_NFAQX_READ(N,F,A : integer; var Data : word;var Q, X : boolean);
begin
  Data:=CAM_NFA_READ(N,F,A);
  Q:=CAM_Q;
  X:=CAM_X;
end;
{***************************************************************************}

procedure CAM_NFAQX_Write(N,F,A: integer; Data : word; var Q, X : boolean);
begin
  CAM_NFA_WRITE(N,F,A,Data);
  Q:=CAM_Q;
  X:=CAM_X;
end;
{***************************************************************************}

procedure CAM_NFA_Write24(N,F,A:integer;Data : longint);
 var naf, b2 : integer;
     b3 : byte;
 begin
  b2:=0;b3:=0;
  b2:=TWordRec(Data).LoWo;
  b3:=TWordRec(Data).HiWo and 255;
	naf:=(A and $0F) shl 5;
  naf:=(naf or (F and $1F)) shl 5;
  naf:=naf or (N and $1F);
  portw[badr02]:=naf;
  portw[badr04]:=b3;
  portw[badr06]:=b2;
end;
{***************************************************************************}

function CAM_NFA_Read24(N,F,A : integer) : longint;
var li : longint;
    naf : integer;
begin
	naf:=(A and $0F) shl 5;
  naf:=(naf or (F and $1F)) shl 5;
  naf:=naf or (N and $1F);
  portw[badr02]:=naf or $8000;
  li:=portw[badr08];
  li:=(li and $ff) shl 16;
  CAM_NFA_Read24:=li + portw[badr0A];
end;
{***************************************************************************}

procedure CAM_NFAQX_READ24(N,F,A : integer; var Data : longint;var Q, X : boolean);
begin
  Data:=CAM_NFA_READ24(N,F,A);
  Q:=CAM_Q;
  X:=CAM_X;
end;
{***************************************************************************}

procedure CAM_NFAQX_Write24(N,F,A: integer;Data : longint; var Q, X : boolean);
begin
  CAM_NFA_WRITE24(N,F,A,Data);
  Q:=CAM_Q;
  X:=CAM_X;
end;
{***************************************************************************}

function CAM_LAM(Station:integer): Boolean;
begin
  CAM_NFA(station,8,0);
  if CAM_Q then CAM_LAM:=true else CAM_LAM:=false;
end;
{***************************************************************************}

function CAM_LAMF: Boolean;
var  i : byte;
begin
  i:=(port[badr12] and 4)shr 2;
  if i=0 then CAM_LAMF:=true else CAM_LAMF:=false;
end;
{***************************************************************************}

procedure CAM_NFA_BlockRead(N,F,A,length : integer;var buffer:datarr);
var i, nfa: integer;
begin
  nfa:=A shl 5;
	nfa:=(nfa or F) shl 5;
	nfa:=nfa or N ;
  portw[BAdr02]:=nfa;
  for i:=0 to length-2 do buffer[i]:=portw[BADR0C];
  buffer[length-1]:=portw[BADR0A];
end;
{***************************************************************************}

procedure broadc_on;
begin
 portw[BAdr18]:=1;
end;
{***************************************************************************}

procedure broadc_off;
begin
 portw[BAdr1A]:=1;
end;

end.
