// ****************************************************************************
// Copyright (C) 2004  Alexandre Suaide
//
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// The Crate Controler Wiener-CC32 is not authorized for use as
// critical component in life support systems without the express
// written approval of ARW Elektronik Germany.
//
// Maintainer(s): Alexandre Suaide (suaide@if.usp.br)
//
// ****************************************************************************
#ifndef __CAMAC_COMMONCODE__
#define __CAMAC_COMMONCODE__

#include "common.h"

/***************************************************************/
static void read_event_now()
{
  static unsigned short ev_buf[MAX_EVENT_SIZE];
  int i = 0;
  iev=0;
  lock=0;
  
  if(USEADACPLUS == 0) ev_read(ev_buf);
  else ev_read_plus(ev_buf);
	if(iev==0) return;
	if(iev==1 && ev_buf[iev-1]==0xFFFF)
	{
	  iev = 0;
		return;
	}
  
  if(ev_buf[iev-1]!=0xFFFF) 
	{
		printk("WARNING: No end of event 0xFFFF detected. Fixing it!\n");
	  ev_buf[iev] = 0xFFFF;
		iev++;
  }
	if(iev>EVENT_SIZE) 
  {
    printk("WARNING: This event has size equal to EVENT_SIZE = %d iev = %d NBUF = %d NEXTW = %d\n",EVENT_SIZE,iev,NBUF,NEXTW);
    iev=EVENT_SIZE;
    ev_buf[iev-1]=0xFFFF;
  }

  if(nwords+iev+4>=MAXSIZE)
  {
    FLAG[NEXTW++]=nwords;
    for(i=nwords;i<MAXSIZE;i++) cc_buf[offset_b+i-nwords] = 0;
    if(NEXTW>=NBUF) NEXTW=0;
    nwords=0;
    offset_b=NEXTW*MAXSIZE;
    if(NBUFFULL<NBUF) NBUFFULL++;
  }
  if(FLAG[NEXTW]==0 && nwords+iev+4<MAXSIZE)
  {
    nwords=nwords+iev;  
    memcpy((void*)&cc_buf[offset_b],(void*)&ev_buf[0],iev*sizeof(unsigned short int));
    offset_b=offset_b+iev;
  }
  else lost++;
  return;
}
/***************************************************************/
static int ev_read(unsigned short int ev_buf[])
{  
  int i,j,looping;
  unsigned short int par,PAT=0,EX=0,TXR=0;
  unsigned int inst;
  int hbit;
  #ifndef C111AT
    int N,A,F;
    int q;
  #endif
  
  #ifndef CAMEH
    unsigned short int CA=0;
    int lw;
  #else
    unsigned int CA=0;
  #endif
 
  looping=1;
  i=0;
  j=0;
  while ((looping == 1) && (i<INSTR_SIZE) )
  {
    hbit=(cc_inst_buf[i] & 0x00f00000);
    if(hbit>0x200000)
    {
      inst = (cc_inst_buf[i] & 0x007f0000);
      par  = cc_inst_buf[i++] & 0xffff;
    }
    else
    {
      inst = (cc_inst_buf[i] & 0x007fc000);
      par  = cc_inst_buf[i++] & 0x3fff;
    }
    if(inst == 0x5C0000) par = par|0x0fff;
    if((inst == 0x100000))
    {
        //NAF for CAMEH
        #ifdef CAMEH
          #include "NAF/NAF_CAMEH_EVO.code"
        #endif

        //NAF for pcicc32
        #ifdef PCICC32
	        #include "NAF/NAF_PCICC32_EVO.code"
        #endif
      
        // NAF for C111AT
        #ifdef C111AT
          #include "NAF/NAF_C111AT_EVO.code"
        #endif           
     }
    else if(inst == 0x580000)
    {
      if(iev<EVENT_SIZE) ev_buf[iev++] = par; 
      else ev_buf[EVENT_SIZE-1] = 0xFFFF; 
    }          /* OUT # */
    else if(inst == 0x590000)
    {
      if(iev<EVENT_SIZE) ev_buf[iev++] = (CA&0x0fff); 
      else ev_buf[EVENT_SIZE-1] = 0xFFFF;
      
    }  /* OUT CA */
    else if(inst == 0x400000)CA = par;                                          /* MOV #,CA */
    else if(inst == 0x410000)PAT = CA;                                          /* MOV CA,PAT */
    else if(inst == 0x420000)CA = PAT;                                          /* MOV PAT,CA */
    else if(inst == 0x200000)looping=0;                                         /* RTRN */
    else if(inst == 0x250000)i = par-1;                                         /* BRU A */
    else if(inst == 0x2A0000)looping=0;                                         /* INIT */
    else if(inst == 0x300000){if((PAT&par) != 0)i++;}                           /* SKIP PAT .ANY.# */
    else if(inst == 0x320000){if((EX&par) != 0)i++;}                            /* SKIP EX .ANY.# */
    else if(inst == 0x330000){if((CA&par) != 0)i++;}                            /* SKIP CA .ANY.# */
    else if(inst == 0x340000){if((PAT&par) == 0)i++;}                           /* SKIP PAT .NONE.# */
    else if(inst == 0x360000){if((EX&par) == 0)i++;}                            /* SKIP EX .NONE.# */
    else if(inst == 0x370000){if((CA&par) == 0)i++;}                            /* SKIP CA .NONE.# */
    else if(inst == 0x3b0000){if(CA<par) i++;}                                  /* SKIP CA .LT.# */
    else if(inst == 0x3f0000){if(CA>par) i++;}                                  /* SKIP CA .GT.# */
    else if(inst == 0x380000){if(PAT<par) i++;}                                 /* SKIP PAT .LT.# */
    else if(inst == 0x3c0000){if(PAT>par) i++;}                                 /* SKIP PAT .GT.# */
    else if(inst == 0x430000)EX = par;                                          /* SETX # */
    else if(inst == 0x440000)EX = EX + par;                                     /* ADDX # */
    else if(inst == 0x450000)CA = CA >> par;                                    /* SHFT # */
    else if(inst == 0x510000)TXR = CA;                                          /* MOV CA,TXR */
    else if(inst == 0x5A0000){if(iev<EVENT_SIZE) ev_buf[iev++] = PAT; else ev_buf[EVENT_SIZE-1] = 0xFFFF;}          /* OUT PAT */
    else if(inst == 0x5C0000){if(iev<EVENT_SIZE) ev_buf[iev++]=(TXR&par); else ev_buf[EVENT_SIZE-1] = 0xFFFF;}      /* MERG #,TXR */
    else if(inst == 0x600000) udelay(par&0x00ff);                               /* DLAY # */
    else if(inst == 0x700000) par = 0;                                          /* CLRB */
    else if(inst == 0x710000) par = 0;                                          /* SETB */
    else if(inst == 0x000000) return iev;                                       /* RETURN */
    else 
    {
      printk(" Invalid instruction \n");
      printk("inst, par: %x %x %x \n",inst,par,cc_inst_buf[i-1]);
    }
  }
  return iev;
}
/***************************************************************/
static int ev_read_plus(unsigned short int ev_buf[])
{
  unsigned int C;
  unsigned int CMD = 0;
  unsigned int ARG = 0;
  unsigned int N = 0;
  unsigned int A = 0;
  unsigned int F = 0;
  unsigned int O = 0;
  unsigned int V1 = 0;
  unsigned int V2 = 0;
  unsigned int looping = 1;
           int i = 0;
           int BRUR = 0;         
  #ifdef CAMEH
    int q;
  #else
    int lw = 0;    
  #endif
  
  
  while ((looping == 1) && (i<INSTR_SIZE))
  {
    C = cc_inst_buf[i];
    i++;
    //////////////////////////////////////////////////////////////////////
    if(C<=0x007FFFFF) // class 0 
    {
      if(C==0x00000000) {/*NOP*/}             // NOP
      else if(C==0x00100000) return iev;      // RTRN
      else if(C==0x00100001) return iev;      // END
      else if(C==0x00100002) return iev;      // INIT
      else if(C==0x00100003)                  // BRUR
      {
        if(BRUR>0) i = BRURADD[BRUR--]+1;
      }
      ///////////////////////////////////////////////////////////////////
      else
      {
        CMD = C & 0xFFFF0000;
        ARG = C & 0x0000FFFF;        
        if(CMD == 0x00400000)                                  // SPB #
        {
          i = ARG - 1;
          if(BRUR<MAXBRUR) BRURADD[BRUR++] = i;
        }
        else if(CMD == 0x00500000) i = ARG -1;                 // BRU #
        else if(CMD == 0x00600000) set_crate_number(ARG);      // SETC
        else if(CMD == 0x00700000) udelay(ARG);                // DLAY
      }
    }
    //////////////////////////////////////////////////////////////////////
    else if(C<=0x00FFFFFF) // class 1
    {
      CMD = C & 0x00F00000;
      if(CMD == 0x00C00000)                                    // FLUSH
      {
        N   = (C & 0x000F0000)>>16;
        ARG = (C & 0x0000FFFF);
        /* NOT IMPLEMENTED YET */        
      }
      else if(CMD == 0x00F00000)                               // OUT & PAR
      {
        REG1= (C & 0x000F0000)>>16;
        ARG = (C & 0x0000FFFF);
        if(REG1>=0x8) // PAR
        {
          if(iev<EVENT_SIZE) // OUT
          {
            REG1 = REG1 & 0x3;
            if(REG1==0) ev_buf[iev++] = ARG | 0x8000;
            else  ev_buf[iev++] = (REG[REG1-1][ARG] & LIMIT[REG1-1]) | 0x8000; 
          }         
        }
        else 
        {
          if(iev<EVENT_SIZE) // OUT
          {
            if(REG1==0) ev_buf[iev++] = ARG;
            else  ev_buf[iev++] = (REG[REG1-1][ARG] & LIMIT[REG1-1]);
          }
          else 
          { 
            ev_buf[EVENT_SIZE-1] = 0xFFFF; 
            return iev;
          }
        }
      }
      else                                                     // NAF
      {
	      REG1 = (C & 0x00300000)>>20;
        R1   = (C & 0x000F0000)>>16;
        ARG  = (C & 0x0000FFFF);
        //NAF for CAMEH
        #ifdef CAMEH
          #include "NAF/NAF_CAMEH_EVP.code"
        #endif

        //NAF for pcicc32
        #ifdef PCICC32
	        #include "NAF/NAF_PCICC32_EVP.code"
        #endif
      
        // NAF for C111AT
        #ifdef C111AT
          #include "NAF/NAF_C111AT_EVP.code"
        #endif      
      }
    }
    //////////////////////////////////////////////////////////////////////
    else if(C<=0x7FFFFFFF) // class 2
    {
      CMD = C & 0x0F000000;
      REG1 = (C & 0x00C00000)>>22;
      REG2 = (C & 0x00300000)>>20;
      R2   = (C & 0x000F0000)>>16;  
      ARG  = (C & 0x0000FFFF); 
      if(REG1!=0) ARG = REG[REG1-1][ARG]; // OR R1
      if(CMD==0x01000000)                                                 // MERG A,B
      {
        REG[REG2-1][R2] = ((ARG & 0xF000) | (REG[REG2-1][R2] & 0x0FFF)) & LIMIT[REG2-1];
        if(iev<EVENT_SIZE) ev_buf[iev++] = REG[REG2-1][ARG];        
      }
      else if(CMD==0x02000000) REG[REG2-1][R2]  = ARG;                    // MOV A,B
      else if(CMD==0x03000000) REG[REG2-1][R2] += ARG;                    // INC B,A
      else if(CMD==0x04000000) REG[REG2-1][R2] -= ARG;                    // DEC B,A
      else if(CMD==0x05000000) REG[REG2-1][R2]  = REG[REG2-1][R2] & ARG;  // AND B,A
      else if(CMD==0x06000000) REG[REG2-1][R2]  = REG[REG2-1][R2] | ARG;  // OR B,A
      else if(CMD==0x07000000) REG[REG2-1][R2]  = !REG[REG2-1][R2];       // NOT B
      else if(CMD==0x08000000) REG[REG2-1][R2]  = REG[REG2-1][R2] >> ARG; // RSH B,A
      else if(CMD==0x09000000) REG[REG2-1][R2]  = REG[REG2-1][R2] << ARG; // LSH B,A
      else if(CMD==0x0A000000) {if(ARG!=0) REG[REG2-1][R2]  = (REG[REG2-1][R2] % ARG);}  // MOD B,A
      else if(CMD==0x0B000000) REG[REG2-1][R2]  = REG[REG2-1][R2] * ARG;  // MUL B,A
      else if(CMD==0x0C000000) {if(ARG!=0) REG[REG2-1][R2]  = (REG[REG2-1][R2] / ARG);}  // DIV B,A
      
      while(REG[REG2-1][R2]>LIMIT[REG2-1]) REG[REG2-1][R2]-=LIMIT[REG2-1];
      while(REG[REG2-1][R2]<0)             REG[REG2-1][R2]+=LIMIT[REG2-1];
    }
    //////////////////////////////////////////////////////////////////////
    else // class 3                                          // SKIP A.OP.B
    {
      O    = (C & 0x70000000)>>28;
      REG1 = (C & 0x0C000000)>>26;
      REG2 = (C & 0x03000000)>>24;
      R1   = (C & 0x00F00000)>>20;  
      R2   = (C & 0x000F0000)>>16;  
      ARG  = (C & 0x0000FFFF); 
      
      if(REG1>0) V1 = REG[REG1-1][R1];
      else V1 = ARG;
      if(REG2>0) V1 = REG[REG2-1][R2];
      else V2 = ARG;
      
      if(O==0)      {if((V1&V2) != 0) i++;}  // ANY
      else if(O==1) {if((V1&V2) == 0) i++;}  // NONE
      else if(O==2) {if(V1>V2)  i++;}        // GT
      else if(O==3) {if(V1<V2)  i++;}        // LT
      else if(O==4) {if(V1>=V2) i++;}        // GE
      else if(O==5) {if(V1<=V2) i++;}        // LE
      else if(O==6) {if(V1!=V2) i++;}        // NE
      else if(O==7) {if(V1==V2) i++;}        // EQ
    }
  }
  return iev;  
}
/***************************************************************/
static int set_crate_number(int I)
{
  #ifdef PCICC32
    if(CRATES[I])
    {
      CURRENT = CRATES[I];
      return 1;
    }
  #endif
  
  #ifdef C111AT
    outw_p(I,GEO_port);
  #endif
  
  #ifdef CAMEH
    int mask=0;
    mask=mask&0x0000ffcf;
    mask=mask|(I<<4);
    outb_p((short) mask,CAM15);    
  #endif
  
  return 0;
}
/***************************************************************/
static ssize_t cc_read (struct file *file, char *buffer, size_t size, loff_t *loff)
{
  int result;
  int max,nbytes,QX;
  int Dat;
  unsigned short int *offset;
  //unsigned short int dummy[MAX_EVENT_SIZE];
 
  /*  Check that we can really write to the buffer   */
  max=size>>1;
  if(max>iev) max=iev;
  if(max>BUFFER_SIZE) return -1;
  if((result = access_ok(VERIFY_WRITE,buffer,size))&&(buffer != NULL)) return result;

  switch (CC_IO_OP)
  {
    case CCREAD16:
		  QX = cami16(NN,AA,FF,&Dat);		  
      memcpy((void*)buffer, &Dat, size);
      return QX;
      break;
    case CCREAD24:
      QX = cami24(NN,AA,FF,&Dat);
      memcpy((void*)buffer, &Dat, size);
      return QX;
      break;
    case CCREADEVNT:
      if(FLAG[NEXTR] > 0)
      {
        offset = (short int *)cc_buf;
        nbytes = FLAG[NEXTR]*2;
        if(nbytes < 1) lock++;
        FLAG[NEXTR]=0;
        if(NBUFFULL>0) NBUFFULL--;
        offset=offset+MAXSIZE*NEXTR;
	      NEXTR++;
        if(NEXTR >= NBUF) NEXTR = 0;
        memcpy((void*)buffer,(void*)offset,nbytes);
        if(lock > 20) 
        {
	        lock=0;
        }
        return nbytes;
      }
      return 0;
      break;
    default:
      return -1;
  }
}
static ssize_t cc_write (struct file *file, const char *buffer, size_t size, loff_t *loff)
{
  int result, Dat=0;
  unsigned short int dummy[MAX_EVENT_SIZE];
  
  if ((result = access_ok(VERIFY_READ,buffer,size))) return result;

  switch (CC_IO_OP)
  {
    case CCWRITE16:
      copy_from_user(&Dat,buffer,2);
      return camo16(NN,AA,FF,Dat);
      break;
    case CCWRITE24:
      copy_from_user(&Dat,buffer,2);
      return camo24(NN,AA,FF,Dat);
      break;
    case CCLOADEH:
      printk("loading EV-STANDARD\n size = %d",(int)size);
      USEADACPLUS = 0;
      lockEVO = 0;
      copy_from_user((int *)cc_inst_buf,buffer,size);
      printk("loading %d instructions \n",(int)size/4);
      iev=0;
      ev_read(dummy);
      lockEVO = 1;
      return 0;
      break;
    case CCLOADEHPLUS:
      printk("loading EV-PLUS\n size = %d",(int)size);
      USEADACPLUS = 1;
      lockEVO = 0;
      copy_from_user((int *)cc_inst_buf,buffer,size);
      printk("loading %d instructions \n",(int)size/4);
      iev=0;
      ev_read_plus(dummy);
      lockEVO = 1;
      return 0;
      break;
    default:
      printk("I dont know what to do with command %d\n",CC_IO_OP);
      return -1;
  }
}
static int cc_ioctl(struct inode * inode, struct file * file,unsigned int cmd,  unsigned long arg)
{
  int crate = MINOR(inode->i_rdev);
  int i,arg1;
  static long idum;
  #ifdef CAMEH
    int lam;
  #endif
  #ifndef PCICC32
    unsigned short int dummy[MAX_EVENT_SIZE];
  #endif
  LIMIT[0] = 0xFFFF; // maximum value for PAT
  LIMIT[1] = 0xFFFF; // maximum value for EX
  LIMIT[2] = 0xFFFF; // maximum value for CA
  arg1=arg;	
  if (crate>=MAX_CRATES) return -ENODEV;
  switch (cmd) 
  {
    case CCMAXSIZE:
      MAXSIZE=arg;
      if(MAXSIZE<MIN_BUFFER_SIZE)  MAXSIZE=MIN_BUFFER_SIZE;
      if(MAXSIZE>MAX_BUFFER_SIZE)  MAXSIZE=MAX_BUFFER_SIZE;
      EVENT_SIZE = MAXSIZE-2;
      if(EVENT_SIZE>MAX_EVENT_SIZE) EVENT_SIZE=MAX_EVENT_SIZE;
      NBUF=BUFFER_SIZE/MAXSIZE;
      if(NBUF>MAX_BUFFER) NBUF=MAX_BUFFER;
      printk(" maxsize,nbuf = %d %d\n",MAXSIZE,NBUF);
      for(i=0;i<BUFFER_SIZE_ARRAY;i++) cc_buf[i] = 0;
      return MAXSIZE;
      break;
    case CCRESET:
      if(arg == ENABLE || arg==0)
      {					
        printk(" Running...\n");
        IRQCOUNT = 0;
        for(i=0;i<BUFFER_SIZE_ARRAY;i++) cc_buf[i] = 0;
        for(i=0;i<MAX_BUFFER;i++) FLAG[i]=0;
        for(i=0;i<10;i++)
        {
          REG[0][i]=0;
          REG[1][i]=0;
          REG[2][i]=0;
        }
        lost=0;
        nwords=0;
        offset_b=0;
        NEXTW=0;
        NEXTR=0;
        NBUFFULL = 0;
        Running = 1;
        #ifdef CAMEH
          arg1=0;
          set_crate_number(crate);
          lam=(inb(CAM8) & 0x007c);
          printk(" lam %d...\n",lam);
          if(lam<=0x007b && USEADACPLUS ==0)ev_read(dummy);
          if(lam<=0x007b && USEADACPLUS ==1)ev_read_plus(dummy);
        #else
          arg1=ENABLE;
        #endif
	idum  = -534234;
        printk("ADACPLUS MODE = %d\n",USEADACPLUS);
        #ifdef C111AT
          if(USEADACPLUS==0) ev_read(dummy);
          else ev_read_plus(dummy);
        #endif
      }
      if(arg==DISABLE || arg==4) 
      {
        printk(" Stopped...lost %d events\n",lost);
        #ifdef CAMEH
          arg1=4;
        #else
          arg1=DISABLE;
        #endif   				    
      }
      #ifdef PCICC32 
        if(IRQCRATE)
        {
          if(arg==DISABLE) camcl(IRQCRATE,4);
          else if(arg==ENABLE) {CURRENT = IRQCRATE; camcl(IRQCRATE,3);} 
        }
      #else
        set_crate_number(crate); 
        camcl ((int) arg1);
      #endif
      return 0;
      break;
    case CCWAITLAM:
      printk("Invalid call: CCWAITLAM \n");
      return -1;
      break;
    case CCGETLOST:
      return lost;
      break;
    case CCGETUSED:
      return NBUFFULL;
      break;
    case CCGETBUF:
      return NBUF;
      break;
    case CCSETCRATE:
      if(arg<MAX_CRATES) set_crate_number(arg);
      return 0;
    default:			
      NN=arg&0x001f;
      FF=(arg&0x03e0)>>5;
      AA=(arg&0x3c00)>>10;			
      CC_IO_OP=cmd;
      return 0;
  }
}
/***************************************************************/
static int camo16(int N, int A, int F,int D)
{
  #ifdef C111AT
	  int NFA = NAFGEN(N,A,F);
    outw_p(NFA,FNA_port);
    outw_p(D,LW_W);
    return inb(R_S) & 3;
	#endif
	#ifdef CAMEH
    outb((short) N,CAM5);
    outb((short) F,CAM4);
    outb((short) A,CAM3);
    outb((short) D,CAM2);
    outb((short) D>>8,CAM1);
    outb((short) N,CAM7);  /*  start camac cycle */
  return inb(CAM8);
	#endif
  #ifdef PCICC32
	  int NAF = MAKE_CC32_ADR(N,A,F);
    writew(D,CURRENT->pUsr+NAF);
    return 0;
	#endif
	
}
/***************************************************************/
static int camo24(int N, int A, int F,int D)
{
  #ifdef C111AT
	  int NFA = NAFGEN(N,A,F);
    int BH;
    BH = (int) ((D >> 16) & 255); 
    outw_p(NFA,FNA_port);
    outw_p(D,LW_W);
    return inb(R_S) & 3;
	#endif
	#ifdef CAMEH
    outb((short) N,CAM5);
    outb((short) F,CAM4);
    outb((short) A,CAM3);
    outb((short) D,CAM2);
    outb((short) D>>8,CAM1);
    D=D>>16;
    outb((short) D,CAM);
    outb((short) N,CAM7);   /* start camac cycle */
    return inb(CAM8);
	#endif
  #ifdef PCICC32
	  int NAF = MAKE_CC32_ADR(N,A,F);
    writew(D,CURRENT->pUsr+NAF);
    return 0;
	#endif
}
/***************************************************************/
static int cami16(int N, int A, int F, int *D)
{
  #ifdef C111AT
	  int NFA = NAFGEN(N,A,F);
    int lw;
    NFA = NFA |0x8000;
    outw_p(NFA,FNA_port);
    lw = (inw_p(LW_R)&0xffff);
    udelay(1);
    *D = lw;
    return (inb_p(R_S) & 3);
  #endif
	#ifdef CAMEH
    int temp;
    outb((short) N,CAM5);
    outb((short) F,CAM4);
    outb((short) A,CAM3);
    outb((short) N,CAM7);  /* start camac cycle */
    temp=(inb(CAM11)&0x000000ff)+((inb(CAM10)<<8)&0x0000ff00);
    *D=(temp&0x0000ffff);
    return inb(CAM8);
	#endif
  #ifdef PCICC32
	  int NAF = MAKE_CC32_ADR(N,A,F);
    *D = readl(CURRENT->pUsr+NAF);
    return 0;
	#endif
}
/***************************************************************/
static int cami24(int N, int A, int F, int *D)
{
  #ifdef C111AT
	  int NFA = NAFGEN(N,A,F);
    unsigned int lw,hb;
    NFA = NFA |0x8000;
    outw_p(NFA,FNA_port);
    hb = inb_p(HB_R)&0x00ff;
    hb = inb_p(HB_R)&0x00ff;
    lw = inw_p(LW_R);
    lw = lw + (hb<<16);
    *D = lw;
    udelay(1);
    return inb_p(R_S) & 3;
	#endif
	#ifdef CAMEH
    int temp;
    outb((short) N,CAM5);
    outb((short) F,CAM4);
    outb((short) A,CAM3);
    outb((short) N,CAM7);   /* start camac cycle */ 
    temp=(inb(CAM11)&0x000000ff)+((inb(CAM10)<<8)&0x0000ff00)+ ((inb(CAM9)<<16)&0x00ff0000);
    *D=temp&0x00ffffff;
    return inb(CAM8);	
	#endif
  #ifdef PCICC32
	  int NAF = MAKE_CC32_ADR(N,A,F);
    *D = readl(CURRENT->pUsr+NAF);
    return 0;
	#endif
}
#endif
