/*
 * fx2-io -- FX2 CAMAC IO via USB
 * For the FX2LP-128 TinyBoard 
 * 
 * Based on usb-fx2-local-examples - by Wolfgang Wieser ] wwieser (a) gmx <*> de [ 
 * 
 * This file may be distributed and/or modified under the terms of the 
 * GNU General Public License version 2 as published by the Free Software 
 * Foundation. (See COPYING.GPL for details.)
 * 
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 *  Registers - CAMAC
 *
 *  PD0-PD7 - R-lines:  PA2=1 (R1-R8), PA1=1 (R9-R16), PA3=1 (R17-R24)
 *  PE0-PE7 - W-lines:  PA2=1 (W1-W8), PA1=1 (W9-W16)
 *
 *  PC0-PC3: A1-A8,  PC4-PC7: F1-F8 PA5: F16
 *
 *  PA0: LAM(input), PA1-PA3: (R/W bytes), PA5: F16, PA6: Z, PA7: S2
 *  PB0: N1, PB1: N2, PB2: N3, PB3: N4  
 */

#define ALLOCATE_EXTERN
#include <fx2regs.h>
#include <camregs.h>

// Read TRM p.15-115 for an explanation on this. 
// A single nop is sufficient for default setup but like that we're on 
// the safe side. 
#define	NOP		_asm nop _endasm
#define	SYNCDELAY	NOP; NOP; NOP; NOP
#define NCHEV 80
static unsigned char AQUIS=0;
static unsigned char nn=0;
static unsigned char len=0;
xdata unsigned char *dest=EP6FIFOBUF+2;
xdata unsigned char *dest0=EP6FIFOBUF;
static unsigned char dd[NCHEV];

static union{
  unsigned char B[2];
  unsigned short int W;
  }CA;

static union{
  unsigned char B[2];
  unsigned short int W;
  }PAT;

void naf(unsigned char, unsigned char, unsigned char);
void nafd(unsigned char, unsigned char, unsigned char);
void rout(unsigned char,unsigned char, unsigned char);
void pat(unsigned char, unsigned char, unsigned char);
void out(unsigned int);
void camcl(unsigned char);
void Z(void);
void event(void);
void init_camac();
void delay(unsigned char);
void nafl(unsigned char, unsigned char);
void nafh(unsigned char, unsigned char);
#include <event-128.c>

void out(unsigned int H)
{
  dd[nn++]=(H&0x00ff);
  dd[nn++]=((H>>8)&0x00ff);
}
#ifdef FX2_CAM128

void naf(unsigned char N, unsigned char A,unsigned char F)
{
unsigned char AF;
AF=(F<<4)|A;
if(F>15) nafh(AF,N); else nafl(AF,N);
}

void nafl(unsigned char af, unsigned char N){ // for F<16 

 IOC=af; //af = FFFFAAAA (F<16) 
 IOB=1<<(N-1); //select N-line: start camac naf
 if(af<0x8f){ // F=0-7 (read function)
  IOA=0x1a; //(R9-R16)
  CA.B[0]=IOD;
  IOA=0x1c; //(R1-R8)
  CA.B[1]=IOD;
  }
 IOA=IOA|0x80; //S2
 NOP; NOP; NOP;
 IOA=0x1e;//IOA&0xdf; // end of S2
 IOB=0; //end of camac cycle
 IOC=0;
 return;
}

void nafh(unsigned char af, unsigned char N){ // for F>=16
 unsigned char IOAD=0;
 IOC=af; //af = FFFFAAAA
 IOA=0x3e; // F16 selected
 IOB=1<<(N-1); //select N-line: start camac naf
 //if(af<0x8f){ //15<F<24) - write function
  IOA=0x3c;
  if(CA.B[1]==0){IOE=CA.B[0]; IOAD=0x2a;};  //BYTE 1 ONLY !!!!!!
  if(CA.B[0]==0){IOE=CA.B[1]; IOAD=0x2c;} //BYTE 2 ONLY !!!!!!
  IOA=IOAD;//Start S1
  NOP; NOP; NOP;
  IOA=0x3c;//End S1
  IOE=0;
  IOA=0x3e;
//  } else  IOB=1<<(N-1); //select N-line: start camac naf
 NOP; NOP;
 IOA=IOA|0x80; //S2
 NOP; NOP; NOP;
 IOA=0x3e;// end of S2
 IOB=0; //end of camac cycle
 IOA=0x1e;
 IOC=0;
 IOE=0;
 return;
}

void nafd(unsigned char N, unsigned char A, unsigned char F){ // for F>=16 dataless

 IOC=(F<<4)|A; //af = FFFFAAAA (F<16) 
 IOB=1<<(N-1); //select N-line: start camac naf
 IOA=IOA|0x80; //S2
 NOP; NOP; NOP;
 IOA=0x1e;//IOA&0xdf; // end of S2
 IOB=0; //end of camac cycle
 IOC=0;

 return;
}

void rout(unsigned char N, unsigned char A, unsigned char F){ // for N=1, F=0 or 2 
 IOC=(F<<4)|A; //af = FFFFAAAA (F<16)
 IOB=1<<(N-1); //select N-line: start camac naf
 IOA=0x1a; //(R1-R8)
 dd[nn++]=IOD;
 IOA=0x1c; //(R9-R16)
 NOP;
 dd[nn++]=IOD;
 IOA=0x1e;
 IOB=0; //end of camac cycle
 IOC=0;
}

void pat(unsigned char N, unsigned char A, unsigned char F){ 
 IOC=(F<<4)|A; //af = FFFFAAAA (F<16)
 IOB=1<<(N-1); //select N-line: start camac naf
 IOA=0x1a; //(R1-R8)
 PAT.B[0]=IOD;
 PAT.B[1]=0;
 NOP;
 IOA=0x1e;
 IOB=0; //end of camac cycle
 IOC=0;
}

void Z(){
IOA=0x5e; // Z - line
NOP;
IOA=0xde; //Z.S2
NOP; NOP; NOP; //160ns
IOA=0x5e; // end of S2
NOP;
IOA=0x1e; // end of Z cycle
}
#endif

#ifdef FX2_ISA
void pat(unsigned char N, unsigned char A, unsigned char F){ 

  OEA=0xff;
  IOA=N; IOC=0x45; IOC=0xff; //outb_p(N,CAM5);
  IOA=F; IOC=0x44; IOC=0xff; //outb_p(F,CAM4);
  IOA=A; IOC=0x43; IOC=0xff; //outb_p(A,CAM3);

    IOA=N; IOC=0x47; IOC=0xff; // outb_p(N,CAM7);  /* start camac cycle */
    OEA=0x00;
    IOC=0x8b; SYNCDELAY; NOP; NOP; PAT.B[0]=IOA; // CA[0]=inb_p(CAM11);
    IOC=0x8a; SYNCDELAY; NOP; NOP; PAT.B[1]=IOA; // CA[1]=inb_p(CAM10); clear chn ID on ADC data

  return;
}

void nafd(unsigned char N, unsigned char A,unsigned char F)
{
  OEA=0xff;
  IOA=N; IOC=0x45; IOC=0xff; //outb_p(N,CAM5);
  IOA=F; IOC=0x44; IOC=0xff; //outb_p(F,CAM4);
  IOA=A; IOC=0x43; IOC=0xff; //outb_p(A,CAM3);
  IOA=N; IOC=0x47; IOC=0xff; // outb_p(N,CAM7);  /* start camac cycle */
}

void rout(unsigned char N, unsigned char A, unsigned char F){

  OEA=0xff;
  IOA=N; IOC=0x45; IOC=0xff; //outb_p(N,CAM5);
  IOA=F; IOC=0x44; IOC=0xff; //outb_p(F,CAM4);
  IOA=A; IOC=0x43; IOC=0xff; //outb_p(A,CAM3);
  IOA=N; IOC=0x47; IOC=0xff; // outb_p(N,CAM7);  /* start camac cycle */
  OEA=0x00;
  IOC=0x8b; NOP; NOP;  dd[nn++]=IOA; // CA[0]=inb_p(CAM11);
  IOC=0x8a; NOP; NOP;  dd[nn++]=(IOA&0x0f); // CA[1]=inb_p(CAM10); clear chn ID on ADC data
  return;
}

void naf(unsigned char N, unsigned char A, unsigned char F){

  OEA=0xff;
  IOA=N; IOC=0x45; IOC=0xff; //outb_p(N,CAM5);
  IOA=F; IOC=0x44; IOC=0xff; //outb_p(F,CAM4);
  IOA=A; IOC=0x43; IOC=0xff; //outb_p(A,CAM3);

  if(F<8){

    IOA=N; IOC=0x47; IOC=0xff; // outb_p(N,CAM7);  /* start camac cycle */
    OEA=0x00;
    IOC=0x8b; SYNCDELAY; NOP; NOP; CA.B[0]=IOA; // CA[0]=inb_p(CAM11);
    IOC=0x8a; SYNCDELAY; NOP; NOP; CA.B[1]=IOA; // CA[1]=inb_p(CAM10); clear chn ID on ADC data
    return;
    }

  if(F>7&F<24){
    IOA=CA.B[0]; IOC=0x42; IOC=0xff;
    IOA=CA.B[1]; IOC=0x41; IOC=0xff;
    IOA=N; IOC=0x47; IOC=0xff; // outb_p(N,CAM7);  /* start camac cycle */
    return;
    }
  if(F>23){
    IOA=N; IOC=0x47; IOC=0xff; // outb_p(N,CAM7);  /* start camac cycle */ 
    return;
    }
}

void Z(void)
{
  camcl(1);
}

/*
camcl:(I)
0x01 - Z
0x02 - C 
0x04 - set I
0x00 - clear I
0x40 - Init
*/
void camcl(unsigned char I){

   IOA=I; NOP; IOC=0x46; NOP; IOC=0xff; // outb_p(I,CAM6);
}
#endif

void delay(unsigned char n) //usecond delay
{
unsigned char i;
for(i=0;i<n;i++)NOP;NOP;NOP;NOP;NOP;NOP;NOP;NOP;NOP;NOP;NOP;NOP; //12NOP=999.8 ns
}

/*
void outb_p(unsigned char DAT, unsigned char PORT)
{
 OEA=0xff; //PORT A output
 IOA=DAT;
 IOC=0x40|PORT;
 IOC=0xff;
 return;
}

unsigned char inb_p(unsigned char PORT)
{
 OEA=0x00; //PORT A input 
 IOC=0x80|PORT;
 IOC=0xff;
 return IOA;
}
*/
#ifdef FX2_ISA
static void Initialize(void)
{
//	CPUCS=0x10;   // 48 MHz, CLKOUT output disabled. 
        CPUCS=0x12; // 24 MHz CLKOUT enabled	
	IFCONFIG=0xc0;  // Internal IFCLK, 48MHz; A,B as normal ports. 
	SYNCDELAY;
	
	REVCTL=0x03;  // See TRM...
	SYNCDELAY;
	
	EP1INCFG=0x00;
	EP1OUTCFG=0xff;
	EP6CFG=0xe2;  // 1110 0010 (bulk IN, 512 bytes, double-buffered)
	SYNCDELAY;

        PORTACFG=0x00; // bits(0-7) IOA
        PORTCCFG=0x00; // bits(0-7) IOC
        PORTECFG=0x00; // bits(0-7) IOE
        OEA=0xff; //PORTA OUT
        OEC=0xff; //PORT C OUT
        OEE=0x00; //PORT E INP
	IOA=0x00;
        IOB=0;
        IOC=0xff;
	FIFORESET = 0x80;  SYNCDELAY;  // NAK all requests from host. 
	FIFORESET = 0x82;  SYNCDELAY;  // Reset individual EP (2,4,6,8)
	FIFORESET = 0x84;  SYNCDELAY;
	FIFORESET = 0x86;  SYNCDELAY;
	FIFORESET = 0x88;  SYNCDELAY;
	FIFORESET = 0x00;  SYNCDELAY;  // Resume normal operation.
	EP1OUTBC=0xff; // arm endpoint 1 for OUT (host->device) transfers
        camcl(0x40);
	camcl(0x01);
	init_camac();

}


static void ProcessEP1Data(void)
{
	xdata const unsigned char *src=EP1OUTBUF;
	unsigned char cc;
	cc=*src;
        IOA=0; IOC=(0x40|CAM15); IOC=0xff; // outb_p(0,CAM15); - Select Crate 0
	if(cc==0x53){  //STOP
         AQUIS=0;
         camcl(0x40);
	 camcl(0x04); //set I
         *dest0=len;
         *(dest0+1)=0;
         SYNCDELAY;  EP6BCH=0;
         SYNCDELAY;  EP6BCL=len;
         SYNCDELAY;
	 dest=EP6FIFOBUF+2;
	 dest0=EP6FIFOBUF;
         len=0;
          }

	if(cc==0x52){  //START
          camcl(0x40); // Init
	  camcl(0x01); //Z
	  camcl(0x00); //clear I
	  init_camac();
          AQUIS=1;
          }
	
	EP1OUTBC=0xff; // re-arm endpoint 1
}
#endif

#ifdef FX2_CAM128
static void Initialize(void)
{
	CPUCS=0x10;   // 48 MHz, CLKOUT output disabled. 
	
	IFCONFIG=0xc0;  // Internal IFCLK, 48MHz; A,B as normal ports. 
	SYNCDELAY;
	
	REVCTL=0x03;  // See TRM...
	SYNCDELAY;
	
	EP1INCFG=0x00;
	EP1OUTCFG=0xff;
	EP6CFG=0xe2;  // 1110 0010 (bulk IN, 512 bytes, double-buffered)
	SYNCDELAY;

        PORTACFG=0x00; // bits(0-7) IOA
        PORTCCFG=0x00; // bits(0-7) IOC
        PORTECFG=0x00; // bits(0-7) IOE
        OEA=0xfe; //PORTA input (LAM bit 0) and output (W bits 1-7)
	OEB=0xff; //PORTB output (W bits 0-7)
        OEC=0xff; //PORT C output (W bits 0-7)	
	OED=0x00; //PORTD input (R)
        OEE=0xff; //PORTE output
	IOA=0x1e;
        IOB=0;
        IOC=0;
        IOE=0;
	FIFORESET = 0x80;  SYNCDELAY;  // NAK all requests from host. 
	FIFORESET = 0x82;  SYNCDELAY;  // Reset individual EP (2,4,6,8)
	FIFORESET = 0x84;  SYNCDELAY;
	FIFORESET = 0x86;  SYNCDELAY;
	FIFORESET = 0x88;  SYNCDELAY;
	FIFORESET = 0x00;  SYNCDELAY;  // Resume normal operation.
	EP1OUTBC=0xff; // arm endpoint 1 for OUT (host->device) transfers

        Z(); // Z on module
	init_camac();
}

static void ProcessEP1Data(void)
{
	xdata const unsigned char *src=EP1OUTBUF;
	unsigned char cc;
	cc=*src;
	if(cc==0x53){  //STOP
         AQUIS=0;
//         Z(); // Z on module
         *dest0=len;
         *(dest0+1)=0;
         SYNCDELAY;  EP6BCH=0;
         SYNCDELAY;  EP6BCL=len;
	 dest=EP6FIFOBUF+2;
	 dest0=EP6FIFOBUF;
         len=0;
          }

	if(cc==0x52){  //START
          Z(); // Z on module
	  init_camac();
          AQUIS=1;
          }
	
	EP1OUTBC=0xff; // re-arm endpoint 1
}
#endif

void main(void)
{
	//unsigned char i;
	Initialize();

	for(;;)
	{
		if (!(EP1OUTCS & (1<<1))) ProcessEP1Data();
		// Wait for the EP6 buffer to become non-full. 
		if(!(EP6CS & (0x08)))
		{
#ifdef FX2_ISA
		if((AQUIS==1)&((IOE&0x20)==0x20)){
#endif
#ifdef FX2_CAM128
		if((AQUIS==1)&((IOA&0x01)==0)){	  
#endif
                  event();
                  }
		}
	}
}

