// ****************************************************************************
// Copyright (C) 2004  Alexandre Suaide
//
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// The Crate Controler Wiener-CC32 is not authorized for use as 
// critical component in life support systems without the express 
// written approval of ARW Elektronik Germany.
//
// Maintainer(s): Alexandre Suaide (suaide@if.usp.br)
//
// ****************************************************************************

#ifndef PCICC32
#define PCICC32

#include "askpci.h"
#include "plxbug.h"
#include "plx9050.h"

#define VERSION_HI 4
#define VERSION_LO 6

#define MAJOR_NO                       26                         /* use dynamic assignment */
#define DEVICE_NAME                    "pcicc32"

#if (VERSION_HI < 3)
  #define ARWVME_DEVICE_ID               0x9050
  #define ARWVME_VENDOR_ID               0x10B5
  #define ARWVME_SUBSYSTEM_ID            0x1167
  #define ARWVME_SUBSYSTEM_VENDOR_ID     0x9050
#else
  #define ARWVME_DEVICE_ID               0x2258
  #define ARWVME_VENDOR_ID               0x10B5
  #define ARWVME_SUBSYSTEM_ID            0x2258
  #define ARWVME_SUBSYSTEM_VENDOR_ID     0x9050
#endif

#define MAKE_CC32_ADR(N, A, F) (u16)((N << 10) + (A << 6) + ((F & 0xf) << 2))

// this is the cc111T and cameh NAF generation definition
#define NFA2NAF(NFA) (u16)(((NFA&0x001f)<<10) + (((NFA&0x3c00)>>10)<<6) + ((((NFA&0x03e0)>>5)&0xf)<<2))

#define ENABLE_PCIADA_IRQS  (u16)0x0049 /* enable PCIADA IRQs            */
#define DISABLE_PCIADA_IRQS (u16)0x0009 /* disable PCIADA IRQs           */
#define ENABLE_ACCESS_AUTOREAD    (u16)0x4182 
#define ENABLE_ACCESS_NOAUTOREAD  (u16)0x4184 
#define DISABLE_ACCESS            (u16)0x4084 

#ifndef NULL
  #define NULL 0
#endif

#define PCICC32_MAGIC ' '

#define PCICC32_IOSTATE _IOR(PCICC32_MAGIC, 1, sizeof(PCICC32_STATUS))                  /* get the status of PCICC32              */
#define PCICC32_IOCNTRL _IO(PCICC32_MAGIC, 2)       			                              /* reset a timeout failure, no parameters */
#define PCICC32_CONTROL_INTERRUPTS _IOWR(PCICC32_MAGIC, 3, sizeof(PCICC32_IRQ_CONTROL)) /* set, clear interrupt enable            */
#define PCICC32_IOSTATE_BLOCKING   _IOR(PCICC32_MAGIC, 4, sizeof(PCICC32_STATUS))       /* calls a blocking IO until interrupt    */
#define PCICC32_AUTOREAD_SWITCH    _IOWR(PCICC32_MAGIC, 4, sizeof(PCICC32_AUTOREAD))    /* switches autoread on or off            */

// the returned cause of the interrupt
#define NOT_MY_INTERRUPT   0
#define PCIADA_INTERRUPT   1
#define CC32_INTERRUPT     2

// the LAM (interrupt) masks
#define LAM_IRQ             0x00FFFFFF  // there was a LAM responible for the timeout
#define CONNECTION_TIMEOUT	0x08000000  // irq raised through a connection timout
#define LAM_BUS_OR          0x10000000  // a LAM-BUS-OR is pending
#define LAM_NOT_OR          0x20000000  // a LAM-NOT-OR is pending
#define LAM_AND_OR          0x40000000  // a LAM-AND-OR is pending
#define LAM_FF              0x80000000  // the LAM-Flip-Flop was set


/*--- TYPEDEFS ----------------------------------------------------------------------------*/
typedef struct
{
	u16 wIndex;                    /* running index of all PCIADAs */
	PCIConfigHeader *pPch;         /* associated PCI configuration */
	u32 pLCR;                      /* base of LCR */
	u32 pUsr;                      /* base of User space */
	u8  bConnected;                /* is it connected ?? */
	u8  cModuleNumber;             /* module number */
	u8  cFPGAVersion;              /* FPGA Version number */
	u16 wInitStep;                 /* counts the done init steps */
	u16 wIrq;                      /* the assigned irq */
	u32 dwInterruptCount;          /* counts the LAM and timeout interrupts */
	u16 wIrqStatus;                /* last cause / status of interrupts */
	wait_queue_head_t event_queue; /* handle interrupt events */
} CC32_DESCRIPTOR;

/*--- PROTOTYPES -------------------------------------------------------------------------*/
static int          request_io_memory(PCIConfigHeader*);
static void         release_io_memory(PCIConfigHeader*);
static int          translate_addresses(CC32_DESCRIPTOR *, PCIConfigHeader *);
static void         un_translate_addresses(CC32_DESCRIPTOR *, PCIConfigHeader *);
int                 test_connection(CC32_DESCRIPTOR *);
int                 get_module_info(CC32_DESCRIPTOR *, u8 *, u8 *);

static int          pcicc32_read_proc(char *, char **, off_t, int, int *, void *);

static int          my_interrupt(u16);
static void         switch_CC32_on(CC32_DESCRIPTOR *);
static void         switch_CC32_off(CC32_DESCRIPTOR *);

char*               tobin(u32);
void                printNAF(CC32_DESCRIPTOR *,int,int,int);

#endif
