/*!\class UsbIO
\author Alexandre A. P. Suaide 

<p>
This class is derived from PelIO and overwrites some of PelIO's methods to
allow connection to a USB system for data acquisition. 
*/
#ifndef UsbIO_HH
#define UsbIO_HH

#include "PelIO.h"
#include "TROOT.h"

class UsbIO : public PelIO
{
  private:    
    bool               mIsUsb;  ///< bool variable to flag if USB is being used
    int                readBlock(); ///< Read event block from USB
      
  public:
                       UsbIO(char* name="UsbIO",char *title="Pellectron IO"); ///< Default constructor
    virtual            ~UsbIO();///< Default destructor
    
    // Input/output methods
    int                openInput(char* file="/dev/c111a0"); ///< Open input stream
    int                closeInput(); ///< Close input stream
    
    // CAMAC related methods
    int                loadEVO(char*); ///< Load Event Handler file into memory
    int                start(); ///< Start USB acquisition
    int                stop(); ///< Stop USB acquisition
    int                getNEventsLost(); ///< Get number of events lost during acquisition
    int                getNUsedBuffers(); ///< Get number of used buffers
    int                getNBuffers(); ///< Get number of buffers
    int                readScalers(int,int*,int=0); ///< Read scalers
    int                clearScalers(int,int=0); ///< Clear scalers
    int                setBufferSize(int); ///<Sets USB buffer size in byets
        
  ClassDef(UsbIO,1)
};
#endif
